@@ -6,33 +6,31 @@
 """Song from MusicXML file."""
 
 ###############################################################################
-# Standard Library imports
+# Imports
 ###############################################################################
 
+# Standard library imports
 import copy
 import pkgutil
-
 from datetime import datetime
-from typing import cast, Optional
+from pathlib import PurePosixPath
 
-###############################################################################
 # Library imports
-###############################################################################
-
-import music21  # type: ignore
-
+import music21
 from mako.template import Template  # type: ignore
 
-###############################################################################
-# Project imports
-###############################################################################
-
-from .channel import Channel
-from .echo import EchoConfig
-from .instrument import DEFAULT_DYN, inst_from_name, InstrumentConfig
-from .reduction import reduce, remove_unused_instruments
-from .shared import CRLF, MusicXmlException
-from .tokens import (
+# Package imports
+from smw_music import __version__
+from smw_music.music_xml.channel import Channel
+from smw_music.music_xml.echo import EchoConfig
+from smw_music.music_xml.instrument import (
+    Dynamics,
+    InstrumentConfig,
+    SampleSource,
+)
+from smw_music.music_xml.reduction import reduce, remove_unused_instruments
+from smw_music.music_xml.shared import CRLF, MusicXmlException
+from smw_music.music_xml.tokens import (
     Annotation,
     CrescDelim,
     Crescendo,
@@ -51,7 +49,6 @@ from .tokens import (
     Triplet,
     Vibrato,
 )
-from .. import __version__
 
 ###############################################################################
 # Private function definitions
@@ -111,7 +108,7 @@ def _get_trems(part: music21.stream.Part) -> list[list[int]]:
 ###############################################################################
 
 
-def _is_crescendo(elem: music21.stream.Stream) -> bool:
+def _is_crescendo(elem: music21.Music21Object) -> bool:
     """
     Test to see if a music21 stream element is a crescendo or diminuendo
     object.
@@ -135,7 +132,7 @@ def _is_crescendo(elem: music21.stream.Stream) -> bool:
 ###############################################################################
 
 
-def _is_line(elem: music21.stream.Stream) -> bool:
+def _is_line(elem: music21.Music21Object) -> bool:
     """
     Test to see if a music21 stream element is a Line object.
 
@@ -155,7 +152,7 @@ def _is_line(elem: music21.stream.Stream) -> bool:
 ###############################################################################
 
 
-def _is_measure(elem: music21.stream.Stream) -> bool:
+def _is_measure(elem: music21.Music21Object) -> bool:
     """
     Test to see if a music21 stream element is a Measure object.
 
@@ -175,7 +172,7 @@ def _is_measure(elem: music21.stream.Stream) -> bool:
 ###############################################################################
 
 
-def _is_slur(elem: music21.stream.Stream) -> bool:
+def _is_slur(elem: music21.Music21Object) -> bool:
     """
     Test to see if a music21 stream element is a Slur object.
 
@@ -195,7 +192,7 @@ def _is_slur(elem: music21.stream.Stream) -> bool:
 ###############################################################################
 
 
-def _is_trill(elem: music21.stream.Stream) -> bool:
+def _is_trill(elem: music21.Music21Object) -> bool:
     return isinstance(elem, music21.expressions.TrillExtension)
 
 
@@ -283,6 +280,11 @@ class Song:
         except music21.converter.ConverterFileException as e:
             raise MusicXmlException(str(e)) from e
 
+        if not isinstance(stream, music21.stream.Score):
+            raise MusicXmlException(
+                f"Can only operate on Scores, not {type(stream)}s"
+            )
+
         for elem in stream.flat:
             if isinstance(elem, music21.metadata.Metadata):
                 metadata["composer"] = elem.composer or "COMPOSER HERE"
@@ -317,32 +319,31 @@ class Song:
     # Private method definitions
     ###########################################################################
 
-    def _collect_instruments(self):
-        instruments: dict[str, set[str]] = {}
+    def _collect_instruments(self) -> None:
+        inst_dyns: dict[str, set[Dynamics]] = {}
         transposes: dict[str, int] = {}
-        inst: str = ""
+
         for channel in self.channels:
             for token in channel.tokens:
                 if isinstance(token, Instrument):
                     inst = token.name
-                    if inst not in instruments:
-                        instruments[inst] = set()
+                    if inst not in inst_dyns:
+                        inst_dyns[inst] = set()
                         transposes[inst] = token.transpose
                 if isinstance(token, Dynamic):
-                    instruments[inst].add(token.level.upper())
+                    inst_dyns[inst].add(Dynamics[token.level.upper()])
                 if isinstance(token, Crescendo):
-                    instruments[inst].add(token.level.upper())
+                    inst_dyns[inst].add(Dynamics[token.target.upper()])
 
-        inst = sorted(instruments)
+        inst_names = sorted(inst_dyns)
 
         self.instruments = [
             InstrumentConfig(
-                x,
-                inst_from_name(x),
-                dynamics_present=instruments[x],
-                transpose=transposes[x],
+                inst,
+                dynamics_present=inst_dyns[inst],
+                transpose=transposes[inst],
             )
-            for x in inst
+            for inst in inst_names
         ]
 
     ###########################################################################
@@ -392,13 +393,19 @@ class Song:
                 # Pick off the instrument transposition
                 transpose = subpart.transposition
                 semitones = transpose.semitones if transpose is not None else 0
+
+                if not isinstance(semitones, int):
+                    raise MusicXmlException(
+                        f"Non-integer transposition {semitones}"
+                    )
+
                 semitones %= 12
 
                 channel_elem.append(Instrument(name, semitones))
             if not isinstance(subpart, music21.stream.Measure):
                 continue
 
-            measure = cast(music21.stream.Measure, subpart)
+            measure = subpart
             note_no = 0
             channel_elem.append(Measure(measure.number))
             if measure.number in sections:
@@ -433,7 +440,11 @@ class Song:
                         triplets = False
 
                 if isinstance(subelem, music21.dynamics.Dynamic):
-                    channel_elem.append(Dynamic.from_music_xml(subelem))
+                    try:
+                        channel_elem.append(Dynamic.from_music_xml(subelem))
+                    except MusicXmlException as e:
+                        msg = f"{e} in measure {measure.number}"
+                        channel_elem.append(Error(msg))
                 if isinstance(
                     subelem, (music21.note.Note, music21.note.Unpitched)
                 ):
@@ -447,10 +458,16 @@ class Song:
                         channel_elem.append(Slur(False))
 
                     # Gross, fix this
-                    if subelem.id in cresc[0]:
+                    subelem_id = subelem.id
+                    if not isinstance(subelem_id, int):
+                        raise MusicXmlException(
+                            f"Non-integer element ID {subelem_id}"
+                        )
+
+                    if subelem_id in cresc[0]:
                         channel_elem.append(
                             CrescDelim(
-                                True, cresc_type[cresc[0].index(subelem.id)]
+                                True, cresc_type[cresc[0].index(subelem_id)]
                             )
                         )
 
@@ -465,10 +482,16 @@ class Song:
                         channel_elem.append(Vibrato(False))
 
                     # Also gross, fix this
-                    if subelem.id in cresc[1]:
+                    subelem_id = subelem.id
+                    if not isinstance(subelem_id, int):
+                        raise MusicXmlException(
+                            f"Non-integer element ID {subelem_id}"
+                        )
+
+                    if subelem_id in cresc[1]:
                         channel_elem.append(
                             CrescDelim(
-                                False, cresc_type[cresc[1].index(subelem.id)]
+                                False, cresc_type[cresc[1].index(subelem_id)]
                             )
                         )
 
@@ -501,7 +524,7 @@ class Song:
         self,
         loop_analysis: bool,
         superloop_analysis: bool,
-    ):
+    ) -> None:
         self._reduced_channels = copy.deepcopy(self.channels)
 
         for n, chan in enumerate(self._reduced_channels):
@@ -515,7 +538,7 @@ class Song:
 
     ###########################################################################
 
-    def _validate(self):
+    def _validate(self) -> None:
         errors = []
         for n, channel in enumerate(self._reduced_channels):
             msgs = channel.check()
@@ -536,8 +559,7 @@ class Song:
         superloop_analysis: bool = True,
         measure_numbers: bool = True,
         include_dt: bool = True,
-        echo_config: Optional[EchoConfig] = None,
-        custom_samples: bool = True,
+        echo_config: EchoConfig | None = None,
         optimize_percussion: bool = True,
     ) -> str:
         """
@@ -557,8 +579,6 @@ class Song:
             True iff current date/time is included in MML
         echo_config: EchoConfig
             Echo configuration
-        custom_samples: bool
-            True iff the custom samples header should be included in the MML
         optimize_percussion: bool
             True iff repeated percussion notes should not repeat their
             instrument
@@ -578,7 +598,45 @@ class Song:
 
         percussion = any(x.percussion for x in self._reduced_channels)
 
-        tmpl = Template(  # nosec - generates a .txt output, no XSS concerns
+        instruments = copy.deepcopy(self.instruments)
+        samples: list[tuple[str, str, int]] = []
+        sample_id = 30
+
+        # TODO: Really shouldn't have this hardcoded here...
+        custom_path = PurePosixPath("custom")
+
+        for inst in instruments:
+            if inst.sample_source == SampleSource.SAMPLEPACK:
+                # TODO: This is pretty blah, song shouldn't rely on pathlib,
+                # see if this can be refactored
+                fname = str(
+                    custom_path / inst.pack_sample[0] / inst.pack_sample[1]
+                )
+                samples.append((fname, inst.brr_str, sample_id))
+                inst.instrument_idx = sample_id
+                sample_id += 1
+            if inst.sample_source == SampleSource.BRR:
+                fname = str(custom_path / inst.brr_fname.name)
+                samples.append((fname, inst.brr_str, sample_id))
+                inst.instrument_idx = sample_id
+                sample_id += 1
+
+        # Overwrite muted/soloed instrument sample numbers
+        solo = any(inst.solo for inst in instruments)
+        mute = any(inst.mute for inst in instruments)
+
+        if solo or mute:
+            samples.append(("EMPTY.brr", "$00 $00 $00 $00 $00", sample_id))
+
+            for inst in instruments:
+                if inst.mute or (solo and not inst.solo):
+                    inst.sample_source = SampleSource.OVERRIDE
+                    inst.instrument_idx = sample_id
+
+            # Not necessary, but we keep it for consistency's sake
+            sample_id += 1
+
+        tmpl = Template(  # nosec B702
             pkgutil.get_data("smw_music", "data/mml.txt")
         )
 
@@ -590,9 +648,9 @@ class Song:
             datetime=build_dt,
             percussion=percussion,
             echo_config=echo_config,
-            instruments=self.instruments,
-            custom_samples=custom_samples,
-            dynamics=list(DEFAULT_DYN.keys()),
+            instruments=instruments,
+            custom_samples=samples,
+            dynamics=list(Dynamics),
         )
 
         rv = rv.replace(" ^", "^")
@@ -601,7 +659,7 @@ class Song:
         # This last bit removes any empty lines at the end (these don't
         # normally show up, but can if the last section in the last staff is
         # empty.
-        return rv.rstrip() + CRLF
+        return str(rv).rstrip() + CRLF
 
     ###########################################################################
 
@@ -613,8 +671,7 @@ class Song:
         superloop_analysis: bool = True,
         measure_numbers: bool = True,
         include_dt: bool = True,
-        echo_config: Optional[EchoConfig] = None,
-        custom_samples: bool = False,
+        echo_config: EchoConfig | None = None,
         optimize_percussion: bool = True,
     ) -> str:
         """
@@ -636,8 +693,6 @@ class Song:
             True iff current date/time is included in MML
         echo_config: EchoConfig
             Echo configuration
-        custom_samples: bool
-            True iff the custom samples header should be included in the MML
         optimize_percussion: bool
             True iff repeated percussion notes should not repeat their
             instrument
@@ -649,7 +704,6 @@ class Song:
             measure_numbers,
             include_dt,
             echo_config,
-            custom_samples,
             optimize_percussion,
         )
 
